from logging import getLogger

from apluggy import PluginManager

from . import plugins, spec


def build_hook() -> PluginManager:
    hook = PluginManager(spec.PROJECT_NAME)
    hook.add_hookspecs(spec)

    plugins.register(hook)

    logger = getLogger(__name__)

    plugin_names = [n for n, p in hook.list_name_plugin() if p]
    logger.info(f'Pluggy project name: {hook.project_name!r}')
    logger.info(f'Loaded plugins: {plugin_names}')

    return hook



from dataclasses import dataclass

from nextline.types import PromptNo, TraceNo


@dataclass
class Command:
    pass


@dataclass
class PdbCommand(Command):
    trace_no: TraceNo
    prompt_no: PromptNo
    command: str



import dataclasses
import datetime
from typing import Dict, Optional

from nextline.plugin.spec import hookimpl
from nextline.spawned import OnEndTrace, OnStartTrace, RunArg
from nextline.types import RunNo, TraceInfo, TraceNo
from nextline.utils.pubsub.broker import PubSub


class TraceInfoRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None
        self._trace_info_map: Dict[TraceNo, TraceInfo] = {}

    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_initialize_run(self, run_arg: RunArg) -> None:
        self._run_no = run_arg.run_no
        self._trace_info_map = {}

    @hookimpl
    async def on_end_run(self) -> None:
        while self._trace_info_map:
            # the process might have been killed.
            _, trace_info = self._trace_info_map.popitem()
            trace_info_end = dataclasses.replace(
                trace_info,
                state='finished',
                ended_at=datetime.datetime.utcnow(),
            )
            await self._registry.publish('trace_info', trace_info_end)
        self._run_no = None

    @hookimpl
    async def on_start_trace(self, event: OnStartTrace) -> None:
        assert self._run_no is not None
        trace_info = TraceInfo(
            run_no=self._run_no,
            trace_no=event.trace_no,
            thread_no=event.thread_no,
            task_no=event.task_no,
            state='running',
            started_at=event.started_at,
        )
        self._trace_info_map[event.trace_no] = trace_info
        await self._registry.publish('trace_info', trace_info)

    @hookimpl
    async def on_end_trace(self, event: OnEndTrace) -> None:
        trace_info = self._trace_info_map.pop(event.trace_no, None)
        if trace_info is None:
            # on_end_run() might have already been called
            return
        trace_info_end = dataclasses.replace(
            trace_info,
            state='finished',
            ended_at=event.ended_at,
        )
        await self._registry.publish('trace_info', trace_info_end)



from logging import getLogger
from typing import Any, AsyncIterator, Callable, Optional

import apluggy
from tblib import pickling_support

from nextline.plugin.spec import hookimpl
from nextline.spawned import Command, RunArg, RunResult
from nextline.utils import ExitedProcess, RunningProcess

from .spawn import run_session

pickling_support.install()


class RunSession:
    @hookimpl
    def init(self, hook: apluggy.PluginManager) -> None:
        self._hook = hook

    @hookimpl
    async def on_initialize_run(self, run_arg: RunArg) -> None:
        self._run_arg = run_arg

    @hookimpl
    @apluggy.asynccontextmanager
    async def run(self) -> AsyncIterator[None]:
        ahook = self._hook.ahook
        async with run_session(self._hook, self._run_arg) as (running, send_command):
            await ahook.on_start_run(running_process=running, send_command=send_command)
            yield
            exited = await running
        if exited.raised:
            logger = getLogger(__name__)
            logger.exception(exited.raised)
        self._run_result = exited.returned or RunResult(ret=None, exc=None)
        await ahook.on_end_run(exited_process=exited)


class Signal:
    @hookimpl
    async def on_start_run(self, running_process: RunningProcess[RunResult]) -> None:
        self._running = running_process

    @hookimpl
    async def interrupt(self) -> None:
        self._running.interrupt()

    @hookimpl
    async def terminate(self) -> None:
        self._running.terminate()

    @hookimpl
    async def kill(self) -> None:
        self._running.kill()


class CommandSender:
    @hookimpl
    async def on_start_run(self, send_command: Callable[[Command], None]) -> None:
        self._send_command = send_command

    @hookimpl
    async def send_command(self, command: Command) -> None:
        self._send_command(command)


class Result:
    @hookimpl
    async def on_end_run(self, exited_process: ExitedProcess[RunResult]) -> None:
        self._run_result = exited_process.returned or RunResult(ret=None, exc=None)

    @hookimpl
    def exception(self) -> Optional[BaseException]:
        return self._run_result.exc

    @hookimpl
    def result(self) -> Any:
        return self._run_result.result()



'''A temporary workaround implementation of non-interactive mode.

The non-interactive mode needs to be properly reimplemented in the trace
function.
'''
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Set

from nextline.utils.pubsub import PubSubItem

if TYPE_CHECKING:
    from nextline import Nextline


class Continuous:
    def __init__(self, nextline: Nextline):
        self._nextline = nextline

        self._pubsub_enabled = PubSubItem[bool]()
        self._tasks: Set[asyncio.Task] = set()

    async def start(self) -> None:
        await self._pubsub_enabled.publish(False)
        self._task = asyncio.create_task(self._monitor_state())

    async def close(self) -> None:
        await self._task
        if self._tasks:
            await asyncio.gather(*self._tasks)
        await self._pubsub_enabled.close()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()

    async def _monitor_state(self):
        async for state in self._nextline.subscribe_state():
            if state == 'initialized' and self._tasks:
                _, pending = await asyncio.wait(
                    self._tasks, timeout=0.001, return_when=asyncio.FIRST_COMPLETED
                )
                self._tasks.clear()
                self._tasks.update(pending)

    async def run_and_continue(self) -> None:
        started = asyncio.Event()
        task = asyncio.create_task(self._run_and_continue(started))
        self._tasks.add(task)
        await started.wait()

    async def run_continue_and_wait(self, started: asyncio.Event):
        await self._run_and_continue(started)

    async def _run_and_continue(self, started: asyncio.Event):
        await self._pubsub_enabled.publish(True)
        try:
            async with self._nextline.run_session():
                # started.set()  # TODO: too early. need to investigate why
                async for prompt in self._nextline.prompts():
                    await self._nextline.send_pdb_command(
                        command='continue',
                        prompt_no=prompt.prompt_no,
                        trace_no=prompt.trace_no,
                    )
                    started.set()
                started.set()  # ensure started is set even if no prompt is received
        finally:
            await self._pubsub_enabled.publish(False)

    @property
    def enabled(self) -> bool:
        return self._pubsub_enabled.latest()

    def subscribe_enabled(self):
        return self._pubsub_enabled.subscribe()



__all__ = ['register']

from apluggy import PluginManager

from .argument import RunArgComposer
from .registrars import (
    PromptInfoRegistrar,
    PromptNoticeRegistrar,
    RunInfoRegistrar,
    RunNoRegistrar,
    ScriptRegistrar,
    StateNameRegistrar,
    StdoutRegistrar,
    TraceInfoRegistrar,
    TraceNumbersRegistrar,
)
from .session import CommandSender, Result, RunSession, Signal


def register(hook: PluginManager) -> None:
    hook.register(StdoutRegistrar)
    hook.register(PromptNoticeRegistrar)
    hook.register(PromptInfoRegistrar)
    hook.register(TraceInfoRegistrar)
    hook.register(TraceNumbersRegistrar)
    hook.register(RunInfoRegistrar)
    hook.register(RunNoRegistrar)
    hook.register(StateNameRegistrar)
    hook.register(ScriptRegistrar)
    hook.register(RunArgComposer)
    hook.register(Result)
    hook.register(Signal)
    hook.register(CommandSender)
    hook.register(RunSession)



from logging import getLogger
from typing import Optional

from apluggy import PluginManager

from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import TraceFunction


class TraceFuncCreator:
    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook
        self._logger = getLogger(__name__)

    @hookimpl
    def create_trace_func(self) -> TraceFunction:
        def _trace_func(frame, event, arg) -> Optional[TraceFunction]:
            try:
                return self._hook.hook.global_trace_func(
                    frame=frame, event=event, arg=arg
                )
            except BaseException:
                self._logger.exception('')
                raise

        return _trace_func


class GlobalTraceFunc:
    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    def global_trace_func(self, frame, event, arg) -> Optional[TraceFunction]:
        if self._hook.hook.filter(trace_args=(frame, event, arg)):
            return None
        self._hook.hook.filtered(trace_args=(frame, event, arg))
        return self._hook.hook.local_trace_func(frame=frame, event=event, arg=arg)



import contextlib
import sys


@contextlib.contextmanager
def disable_trace():
    '''Remove the system trace function temporarily.
    
    Example:
    
    >>> with disable_trace():
    ...     # code without tracing
    ...     pass

    '''
    trace = sys.gettrace()
    sys.settrace(None)
    try:
        yield
    finally:
        sys.settrace(trace)



from __future__ import annotations

import os
from types import FrameType
from typing import Any, Callable, ContextManager, Dict, Optional

from .types import TraceFunction


def WithContext(
    trace: TraceFunction, context: Callable[[FrameType, str, Any], ContextManager[None]]
) -> TraceFunction:
    '''Return a new trace func that calls the given trace func in the context.'''

    def _create_local_trace() -> TraceFunction:
        next_trace: TraceFunction | None = trace

        def _local_trace(frame, event, arg) -> Optional[TraceFunction]:
            nonlocal next_trace
            assert next_trace
            with context(frame, event, arg):
                if next_trace := next_trace(frame, event, arg):
                    return _local_trace
                return None

        return _local_trace

    def _global_trace(frame: FrameType, event, arg) -> Optional[TraceFunction]:
        return _create_local_trace()(frame, event, arg)

    return _global_trace


def ToCanonicPath() -> Callable[[str], str]:
    # Based on Bdb.canonic()
    # https://github.com/python/cpython/blob/v3.10.5/Lib/bdb.py#L39-L54

    cache: Dict[str, str] = {}

    def _to_canonic_path(filename: str) -> str:
        if filename == "<" + filename[1:-1] + ">":
            return filename
        canonic = cache.get(filename)
        if not canonic:
            canonic = os.path.abspath(filename)
            canonic = os.path.normcase(canonic)
            cache[filename] = canonic
        return canonic

    return _to_canonic_path


to_canonic_path = ToCanonicPath()



__all__ = [
    'PromptInfoRegistrar',
    'PromptNoticeRegistrar',
    'RunInfoRegistrar',
    'RunNoRegistrar',
    'ScriptRegistrar',
    'StateNameRegistrar',
    'StdoutRegistrar',
    'TraceInfoRegistrar',
    'TraceNumbersRegistrar',
]

from .prompt_info import PromptInfoRegistrar
from .prompt_notice import PromptNoticeRegistrar
from .run_info import RunInfoRegistrar
from .run_no import RunNoRegistrar
from .script import ScriptRegistrar
from .state_name import StateNameRegistrar
from .stdout import StdoutRegistrar
from .trace_info import TraceInfoRegistrar
from .trace_nos import TraceNumbersRegistrar



from typing import Optional

from apluggy import PluginManager

from nextline.count import RunNoCounter
from nextline.plugin.spec import hookimpl
from nextline.spawned import RunArg, Statement
from nextline.utils.pubsub.broker import PubSub

SCRIPT_FILE_NAME = "<string>"


class RunArgComposer:
    @hookimpl
    def init(
        self,
        hook: PluginManager,
        registry: PubSub,
        run_no_start_from: int,
        statement: Statement,
    ) -> None:
        self._hook = hook
        self._registry = registry
        self._run_no_count = RunNoCounter(run_no_start_from)
        self._statement = statement
        self._filename = SCRIPT_FILE_NAME

    @hookimpl
    async def start(self) -> None:
        await self._hook.ahook.on_change_script(
            script=self._statement,
            filename=self._filename,
        )

    @hookimpl
    async def reset(
        self, run_no_start_from: Optional[int], statement: Optional[Statement]
    ) -> None:
        if statement is not None:
            self._statement = statement
            await self._hook.ahook.on_change_script(
                script=self._statement,
                filename=self._filename,
            )
        if run_no_start_from is not None:
            self._run_no_count = RunNoCounter(run_no_start_from)

    @hookimpl
    def compose_run_arg(self) -> RunArg:
        run_arg = RunArg(
            run_no=self._run_no_count(),
            statement=self._statement,
            filename=self._filename,
        )
        return run_arg
